@@ -1,7 +1,5 @@
-# -*- coding: utf-8 -*-
-#
 # This file is part of the python-chess library.
-# Copyright (C) 2012-2019 Niklas Fiekas <niklas.fiekas@backscattering.de>
+# Copyright (C) 2012-2020 Niklas Fiekas <niklas.fiekas@backscattering.de>
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -17,6 +15,7 @@
 # along with this program. If not, see <http://www.gnu.org/licenses/>.
 
 import collections
+import math
 import mmap
 import os
 import re
@@ -25,6 +24,9 @@ import threading
 
 import chess
 
+from types import TracebackType
+from typing import Deque, Dict, Iterable, Iterator, List, MutableMapping, Optional, Tuple, Type, TypeVar, Union
+
 
 UINT64_BE = struct.Struct(">Q")
 UINT32 = struct.Struct("<I")
@@ -46,10 +48,10 @@ TRIANGLE = [
 
 INVTRIANGLE = [1, 2, 3, 10, 11, 19, 0, 9, 18, 27]
 
-def offdiag(square):
+def offdiag(square: chess.Square) -> int:
     return chess.square_rank(square) - chess.square_file(square)
 
-def flipdiag(square):
+def flipdiag(square: chess.Square) -> chess.Square:
     return ((square >> 3) | (square << 3)) & 63
 
 LOWER = [
@@ -86,7 +88,7 @@ FLAP = [
 ]
 
 PTWIST = [
-     0, 0,   0,  0,  0,  0,  0,  0,
+     0,  0,  0,  0,  0,  0,  0,  0,
     47, 35, 23, 11, 10, 22, 34, 46,
     45, 33, 21,  9,  8, 20, 32, 44,
     43, 31, 19,  7,  6, 18, 30, 42,
@@ -286,13 +288,13 @@ PP_IDX = [[
      -1,  -1,  -1,  -1, 276,  -1,  -1,  -1,
      -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,
      -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,
-     -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1
+     -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,
 ]]
 
-def test45(sq):
-    return chess.BB_SQUARES[sq] & (chess.BB_A5 | chess.BB_A6 | chess.BB_A7 |
-                                   chess.BB_B5 | chess.BB_B6 |
-                                   chess.BB_C5)
+def test45(sq: chess.Square) -> bool:
+    return bool(chess.BB_SQUARES[sq] & (chess.BB_A5 | chess.BB_A6 | chess.BB_A7 |
+                                        chess.BB_B5 | chess.BB_B6 |
+                                        chess.BB_C5))
 
 MTWIST = [
     15, 63, 55, 47, 40, 48, 56, 12,
@@ -305,16 +307,11 @@ MTWIST = [
     14, 60, 52, 44, 43, 51, 59, 13,
 ]
 
-BINOMIAL = []
-for i in range(5):
-    BINOMIAL.append([])
-    for j in range(64):
-        f = j
-        l = 1
-        for k in range(1, i + 1):
-            f *= j - k
-            l *= k + 1
-        BINOMIAL[i].append(f // l)
+def binom(x: int, y: int) -> int:
+    try:
+        return math.factorial(x) // math.factorial(y) // math.factorial(x - y)
+    except ValueError:
+        return 0
 
 PAWNIDX = [[0 for _ in range(24)] for _ in range(5)]
 
@@ -326,28 +323,28 @@ for i in range(5):
     s = 0
     while j < 6:
         PAWNIDX[i][j] = s
-        s += 1 if i == 0 else BINOMIAL[i - 1][PTWIST[INVFLAP[j]]]
+        s += 1 if i == 0 else binom(PTWIST[INVFLAP[j]], i)
         j += 1
     PFACTOR[i][0] = s
 
     s = 0
     while j < 12:
         PAWNIDX[i][j] = s
-        s += 1 if i == 0 else BINOMIAL[i - 1][PTWIST[INVFLAP[j]]]
+        s += 1 if i == 0 else binom(PTWIST[INVFLAP[j]], i)
         j += 1
     PFACTOR[i][1] = s
 
     s = 0
     while j < 18:
         PAWNIDX[i][j] = s
-        s += 1 if i == 0 else BINOMIAL[i - 1][PTWIST[INVFLAP[j]]]
+        s += 1 if i == 0 else binom(PTWIST[INVFLAP[j]], i)
         j += 1
     PFACTOR[i][2] = s
 
     s = 0
     while j < 24:
         PAWNIDX[i][j] = s
-        s += 1 if i == 0 else BINOMIAL[i - 1][PTWIST[INVFLAP[j]]]
+        s += 1 if i == 0 else binom(PTWIST[INVFLAP[j]], i)
         j += 1
     PFACTOR[i][3] = s
 
@@ -359,7 +356,7 @@ for i in range(5):
     s = 0
     for j in range(10):
         MULTIDX[i][j] = s
-        s += 1 if i == 0 else BINOMIAL[i - 1][MTWIST[INVTRIANGLE[j]]]
+        s += 1 if i == 0 else binom(MTWIST[INVTRIANGLE[j]], i)
     MFACTOR[i] = s
 
 WDL_TO_MAP = [1, 3, 0, 2, 0]
@@ -373,11 +370,11 @@ PCHR = ["K", "Q", "R", "B", "N", "P"]
 TABLENAME_REGEX = re.compile(r"^[KQRBNP]+v[KQRBNP]+\Z")
 
 
-def is_table_name(name):
-    return len(name) <= 7 + 1 and TABLENAME_REGEX.match(name) and normalize_tablename(name) == name
+def is_table_name(name: str) -> bool:
+    return len(name) <= 7 + 1 and bool(TABLENAME_REGEX.match(name)) and normalize_tablename(name) == name
 
 
-def tablenames(*, one_king=True, piece_count=6):
+def tablenames(*, one_king: bool = True, piece_count: int = 6) -> Iterator[str]:
     first = "K" if one_king else "P"
 
     targets = []
@@ -392,7 +389,7 @@ def tablenames(*, one_king=True, piece_count=6):
     return all_dependencies(targets, one_king=one_king)
 
 
-def normalize_tablename(name, *, mirror=False):
+def normalize_tablename(name: str, *, mirror: bool = False) -> str:
     w, b = name.split("v", 1)
     w = "".join(sorted(w, key=PCHR.index))
     b = "".join(sorted(b, key=PCHR.index))
@@ -402,7 +399,7 @@ def normalize_tablename(name, *, mirror=False):
         return w + "v" + b
 
 
-def _dependencies(target, *, one_king=True):
+def _dependencies(target: str, *, one_king: bool = True) -> Iterator[str]:
     w, b = target.split("v", 1)
 
     for p in PCHR:
@@ -422,7 +419,7 @@ def _dependencies(target, *, one_king=True):
             yield normalize_tablename(w + "v" + b.replace(p, "", 1))
 
 
-def dependencies(target, *, one_king=True):
+def dependencies(target: str, *, one_king: bool = True) -> Iterator[str]:
     closed = set()
     if one_king:
         closed.add("KvK")
@@ -433,7 +430,7 @@ def dependencies(target, *, one_king=True):
             closed.add(dependency)
 
 
-def all_dependencies(targets, *, one_king=True):
+def all_dependencies(targets: Iterable[str], *, one_king: bool = True) -> Iterator[str]:
     closed = set()
     if one_king:
         closed.add("KvK")
@@ -451,7 +448,7 @@ def all_dependencies(targets, *, one_king=True):
         open_list.extend(_dependencies(name, one_king=one_king))
 
 
-def calc_key(board, *, mirror=False):
+def calc_key(board: chess.BaseBoard, *, mirror: bool = False) -> str:
     w = board.occupied_co[chess.WHITE ^ mirror]
     b = board.occupied_co[chess.BLACK ^ mirror]
 
@@ -472,7 +469,7 @@ def calc_key(board, *, mirror=False):
     ])
 
 
-def recalc_key(pieces, *, mirror=False):
+def recalc_key(pieces: List[chess.PieceType], *, mirror: bool = False) -> str:
     # Some endgames are stored with a different key than their filename
     # indicates: http://talkchess.com/forum/viewtopic.php?p=695509#695509
 
@@ -496,7 +493,7 @@ def recalc_key(pieces, *, mirror=False):
     ])
 
 
-def subfactor(k, n):
+def subfactor(k: int, n: int) -> int:
     f = n
     l = 1
 
@@ -507,7 +504,7 @@ def subfactor(k, n):
     return f // l
 
 
-def dtz_before_zeroing(wdl):
+def dtz_before_zeroing(wdl: int) -> int:
     return ((wdl > 0) - (wdl < 0)) * (1 if abs(wdl) == 2 else 101)
 
 
@@ -517,45 +514,46 @@ class MissingTableError(KeyError):
 
 
 class PairsData:
-    def __init__(self):
-        self.indextable = None
-        self.sizetable = None
-        self.data = None
-        self.offset = None
-        self.symlen = None
-        self.sympat = None
-        self.blocksize = None
-        self.idxbits = None
-        self.min_len = None
-        self.base = None
+    indextable: int
+    sizetable: int
+    data: int
+    offset: int
+    symlen: List[int]
+    sympat: int
+    blocksize: int
+    idxbits: int
+    min_len: int
+    base: List[int]
 
 
 class PawnFileData:
-    def __init__(self):
-        self.precomp = {}
-        self.factor = {}
-        self.pieces = {}
-        self.norm = {}
+    def __init__(self) -> None:
+        self.precomp: Dict[int, PairsData] = {}
+        self.factor: Dict[int, List[int]] = {}
+        self.pieces: Dict[int, List[int]] = {}
+        self.norm: Dict[int, List[int]] = {}
 
 
 class PawnFileDataDtz:
-    def __init__(self):
-        self.precomp = None
-        self.factor = None
-        self.pieces = None
-        self.norm = None
+    precomp: PairsData
+    factor: List[int]
+    pieces: List[int]
+    norm: List[int]
 
 
+TableT = TypeVar("TableT", bound="Table")
+
 class Table:
+    size: List[int]
 
-    def __init__(self, path, *, variant=chess.Board):
+    def __init__(self, path: str, *, variant: Type[chess.Board] = chess.Board) -> None:
         self.path = path
         self.variant = variant
 
         self.write_lock = threading.RLock()
         self.initialized = False
-        self.fd = None
-        self.data = None
+        self.fd: Optional[int] = None
+        self.data: Optional[mmap.mmap] = None
 
         self.read_condition = threading.Condition()
         self.read_count = 0
@@ -597,7 +595,7 @@ class Table:
                             j = white_part.count(piece_type)
                         self.enc_type = 1 + j
 
-    def init_mmap(self):
+    def init_mmap(self) -> None:
         with self.write_lock:
             # Open fd.
             if self.fd is None:
@@ -605,12 +603,23 @@ class Table:
 
             # Open mmap.
             if self.data is None:
-                self.data = mmap.mmap(self.fd, 0, access=mmap.ACCESS_READ)
+                data = mmap.mmap(self.fd, 0, access=mmap.ACCESS_READ)
+                if data.size() % 64 != 16:
+                    raise IOError(f"invalid file size: ensure {self.path!r} is a valid syzygy tablebase file")
+                self.data = data
 
-    def check_magic(self, magic):
-        return self.data[:min(len(self.data), len(magic))] == magic
+                try:
+                    # Python 3.8
+                    self.data.madvise(mmap.MADV_RANDOM)
+                except AttributeError:
+                    pass
 
-    def setup_pairs(self, data_ptr, tb_size, size_idx, wdl):
+    def check_magic(self, magic: Optional[bytes], pawnless_magic: Optional[bytes]) -> None:
+        valid_magics = [magic, self.has_pawns and pawnless_magic]
+        if self.data[:min(4, len(self.data))] not in valid_magics:
+            raise IOError(f"invalid magic header: ensure {self.path!r} is a valid syzygy tablebase file")
+
+    def setup_pairs(self, data_ptr: int, tb_size: int, size_idx: int, wdl: int) -> PairsData:
         d = PairsData()
 
         self._flags = self.data[data_ptr]
@@ -665,7 +674,7 @@ class Table:
 
         return d
 
-    def set_norm_piece(self, norm, pieces):
+    def set_norm_piece(self, norm: List[int], pieces) -> None:
         if self.enc_type == 0:
             norm[0] = 3
         elif self.enc_type == 2:
@@ -681,7 +690,7 @@ class Table:
                 j += 1
             i += norm[i]
 
-    def calc_factors_piece(self, factor, order, norm):
+    def calc_factors_piece(self, factor: List[int], order: int, norm: List[int]) -> int:
         if not self.variant.connected_kings:
             PIVFAC = [31332, 28056, 462]
         else:
@@ -708,7 +717,7 @@ class Table:
 
         return f
 
-    def calc_factors_pawn(self, factor, order, order2, norm, f):
+    def calc_factors_pawn(self, factor: List[int], order: int, order2: int, norm: List[int], f: int) -> int:
         i = norm[0]
         if order2 < 0x0f:
             i += norm[i]
@@ -732,7 +741,7 @@ class Table:
 
         return fac
 
-    def set_norm_pawn(self, norm, pieces):
+    def set_norm_pawn(self, norm: List[int], pieces) -> None:
         norm[0] = self.pawns[0]
         if self.pawns[1]:
             norm[self.pawns[0]] = self.pawns[1]
@@ -745,7 +754,7 @@ class Table:
                 j += 1
             i += norm[i]
 
-    def calc_symlen(self, d, s, tmp):
+    def calc_symlen(self, d: PairsData, s: int, tmp: List[int]) -> None:
         w = d.sympat + 3 * s
         s2 = (self.data[w + 2] << 4) | (self.data[w + 1] >> 4)
         if s2 == 0x0fff:
@@ -759,14 +768,14 @@ class Table:
             d.symlen[s] = d.symlen[s1] + d.symlen[s2] + 1
         tmp[s] = 1
 
-    def pawn_file(self, pos):
+    def pawn_file(self, pos: List[chess.Square]) -> int:
         for i in range(1, self.pawns[0]):
             if FLAP[pos[0]] > FLAP[pos[i]]:
                 pos[0], pos[i] = pos[i], pos[0]
 
         return FILE_TO_FILE[pos[0] & 0x07]
 
-    def encode_piece(self, norm, pos, factor):
+    def encode_piece(self, norm: List[int], pos: List[chess.Square], factor: List[int]) -> int:
         n = self.num
 
         if self.enc_type < 3:
@@ -851,7 +860,7 @@ class Table:
             idx = MULTIDX[norm[0] - 1][TRIANGLE[pos[0]]]
             i = 1
             while i < norm[0]:
-                idx += BINOMIAL[i - 1][MTWIST[pos[i]]]
+                idx += binom(MTWIST[pos[i]], i)
                 i += 1
 
         idx *= factor[0]
@@ -872,14 +881,14 @@ class Table:
                 j = 0
                 for l in range(i):
                     j += int(p > pos[l])
-                s += BINOMIAL[m - i][p - j]
+                s += binom(p - j, m - i + 1)
 
             idx += s * factor[i]
             i += t
 
         return idx
 
-    def encode_pawn(self, norm, pos, factor):
+    def encode_pawn(self, norm: List[int], pos: List[chess.Square], factor: List[int]) -> int:
         n = self.num
 
         if pos[0] & 0x04:
@@ -894,7 +903,7 @@ class Table:
         t = self.pawns[0] - 1
         idx = PAWNIDX[t][FLAP[pos[0]]]
         for i in range(t, 0, -1):
-            idx += BINOMIAL[t - i][PTWIST[pos[i]]]
+            idx += binom(PTWIST[pos[i]], t - i + 1)
         idx *= factor[0]
 
         # Remaining pawns.
@@ -911,7 +920,7 @@ class Table:
                 j = 0
                 for k in range(i):
                     j += int(p > pos[k])
-                s += BINOMIAL[m - i][p - j - 8]
+                s += binom(p - j - 8, m - i + 1)
             idx += s * factor[i]
             i = t
 
@@ -928,14 +937,14 @@ class Table:
                 j = 0
                 for k in range(i):
                     j += int(p > pos[k])
-                s += BINOMIAL[m - i][p - j]
+                s += binom(p - j, m - i + 1)
 
             idx += s * factor[i]
             i += t
 
         return idx
 
-    def decompress_pairs(self, d, idx):
+    def decompress_pairs(self, d: PairsData, idx: int) -> int:
         if not d.idxbits:
             return d.min_len
 
@@ -1000,19 +1009,19 @@ class Table:
         else:
             return self.data[w]
 
-    def read_uint64_be(self, data_ptr):
+    def read_uint64_be(self, data_ptr: int) -> int:
         return UINT64_BE.unpack_from(self.data, data_ptr)[0]
 
-    def read_uint32(self, data_ptr):
+    def read_uint32(self, data_ptr: int) -> int:
         return UINT32.unpack_from(self.data, data_ptr)[0]
 
-    def read_uint32_be(self, data_ptr):
+    def read_uint32_be(self, data_ptr: int) -> int:
         return UINT32_BE.unpack_from(self.data, data_ptr)[0]
 
-    def read_uint16(self, data_ptr):
+    def read_uint16(self, data_ptr: int) -> int:
         return UINT16.unpack_from(self.data, data_ptr)[0]
 
-    def close(self):
+    def close(self) -> None:
         with self.write_lock:
             with self.read_condition:
                 while self.read_count > 0:
@@ -1026,31 +1035,32 @@ class Table:
                     os.close(self.fd)
                     self.fd = None
 
-    def __enter__(self):
+    def __enter__(self: TableT) -> TableT:
         return self
 
-    def __exit__(self, exc_type, exc_value, traceback):
+    def __exit__(self, exc_type: Optional[Type[BaseException]], exc_value: Optional[BaseException], traceback: Optional[TracebackType]) -> None:
         self.close()
 
 
 class WdlTable(Table):
+    _next: int
+    _flags: int
 
-    def init_table_wdl(self):
+    def init_table_wdl(self) -> None:
         with self.write_lock:
             self.init_mmap()
 
             if self.initialized:
                 return
 
-            if not self.check_magic(self.variant.tbw_magic) and (self.has_pawns or self.variant.pawnless_tbw_magic is None or not self.check_magic(self.variant.pawnless_tbw_magic)):
-                raise IOError("invalid magic header: ensure {!r} is a valid syzygy tablebase file".format(self.path))
+            self.check_magic(self.variant.tbw_magic, self.variant.pawnless_tbw_magic)
 
             self.tb_size = [0 for _ in range(8)]
             self.size = [0 for _ in range(8 * 3)]
 
             # Used if there are only pieces.
-            self.precomp = {}
-            self.pieces = {}
+            self.precomp: Dict[int, PairsData] = {}
+            self.pieces: Dict[int, List[int]] = {}
 
             self.factor = {0: [0 for _ in range(TBPIECES)],
                            1: [0 for _ in range(TBPIECES)]}
@@ -1061,8 +1071,6 @@ class WdlTable(Table):
             # Used if there are pawns.
             self.files = [PawnFileData() for _ in range(4)]
 
-            self._next = None
-            self._flags = None
             self.flags = None
 
             split = self.data[4] & 0x01
@@ -1080,8 +1088,6 @@ class WdlTable(Table):
                 if split:
                     self.precomp[1] = self.setup_pairs(data_ptr, self.tb_size[1], 3, True)
                     data_ptr = self._next
-                else:
-                    self.precomp[1] = None
 
                 self.precomp[0].indextable = data_ptr
                 data_ptr += self.size[0]
@@ -1117,8 +1123,6 @@ class WdlTable(Table):
                     if split:
                         self.files[f].precomp[1] = self.setup_pairs(data_ptr, self.tb_size[2 * f + 1], 6 * f + 3, True)
                         data_ptr = self._next
-                    else:
-                        self.files[f].precomp[1] = None
 
                 for f in range(files):
                     self.files[f].precomp[0].indextable = data_ptr
@@ -1145,7 +1149,7 @@ class WdlTable(Table):
 
             self.initialized = True
 
-    def setup_pieces_pawn(self, p_data, p_tb_size, f):
+    def setup_pieces_pawn(self, p_data: int, p_tb_size: int, f: int) -> None:
         j = 1 + int(self.pawns[1] > 0)
         order = self.data[p_data] & 0x0f
         order2 = self.data[p_data + 1] & 0x0f if self.pawns[1] else 0x0f
@@ -1163,7 +1167,7 @@ class WdlTable(Table):
         self.files[f].factor[1] = [0 for _ in range(TBPIECES)]
         self.tb_size[p_tb_size + 1] = self.calc_factors_pawn(self.files[f].factor[1], order, order2, self.files[f].norm[1], f)
 
-    def setup_pieces_piece(self, p_data):
+    def setup_pieces_piece(self, p_data: int) -> None:
         self.pieces[0] = [self.data[p_data + i + 1] & 0x0f for i in range(self.num)]
         order = self.data[p_data] & 0x0f
         self.set_norm_piece(self.norm[0], self.pieces[0])
@@ -1174,7 +1178,7 @@ class WdlTable(Table):
         self.set_norm_piece(self.norm[1], self.pieces[1])
         self.tb_size[1] = self.calc_factors_piece(self.factor[1], order, self.norm[1])
 
-    def probe_wdl_table(self, board):
+    def probe_wdl_table(self, board: chess.Board) -> int:
         try:
             with self.read_condition:
                 self.read_count += 1
@@ -1184,7 +1188,7 @@ class WdlTable(Table):
                 self.read_count -= 1
                 self.read_condition.notify()
 
-    def _probe_wdl_table(self, board):
+    def _probe_wdl_table(self, board: chess.Board) -> int:
         self.init_table_wdl()
 
         key = calc_key(board)
@@ -1247,15 +1251,14 @@ class WdlTable(Table):
 
 class DtzTable(Table):
 
-    def init_table_dtz(self):
+    def init_table_dtz(self) -> None:
         with self.write_lock:
             self.init_mmap()
 
             if self.initialized:
                 return
 
-            if not self.check_magic(self.variant.tbz_magic) and (self.has_pawns or self.variant.pawnless_tbz_magic is None or not self.check_magic(self.variant.pawnless_tbz_magic)):
-                raise IOError("invalid magic header: ensure {!r} is a valid syzygy tablebase file".format(self.path))
+            self.check_magic(self.variant.tbz_magic, self.variant.pawnless_tbz_magic)
 
             self.factor = [0 for _ in range(TBPIECES)]
             self.norm = [0 for _ in range(self.num)]
@@ -1345,7 +1348,7 @@ class DtzTable(Table):
 
             self.initialized = True
 
-    def probe_dtz_table(self, board, wdl):
+    def probe_dtz_table(self, board: chess.Board, wdl: int) -> Tuple[int, int]:
         try:
             with self.read_condition:
                 self.read_count += 1
@@ -1355,7 +1358,7 @@ class DtzTable(Table):
                 self.read_count -= 1
                 self.read_condition.notify()
 
-    def _probe_dtz_table(self, board, wdl):
+    def _probe_dtz_table(self, board: chess.Board, wdl: int) -> Tuple[int, int]:
         self.init_table_dtz()
 
         key = calc_key(board)
@@ -1439,13 +1442,13 @@ class DtzTable(Table):
 
         return res, 1
 
-    def setup_pieces_piece_dtz(self, p_data, p_tb_size):
+    def setup_pieces_piece_dtz(self, p_data: int, p_tb_size: int) -> None:
         self.pieces = [self.data[p_data + i + 1] & 0x0f for i in range(self.num)]
         order = self.data[p_data] & 0x0f
         self.set_norm_piece(self.norm, self.pieces)
         self.tb_size[p_tb_size] = self.calc_factors_piece(self.factor, order, self.norm)
 
-    def setup_pieces_pawn_dtz(self, p_data, p_tb_size, f):
+    def setup_pieces_pawn_dtz(self, p_data: int, p_tb_size: int, f: int) -> None:
         j = 1 + int(self.pawns[1] > 0)
         order = self.data[p_data] & 0x0f
         order2 = self.data[p_data + 1] & 0x0f if self.pawns[1] else 0x0f
@@ -1466,17 +1469,17 @@ class Tablebase:
     descriptors at any given time. The least recently used tables are closed,
     if nescessary.
     """
-    def __init__(self, *, max_fds=128, VariantBoard=chess.Board):
+    def __init__(self, *, max_fds: Optional[int] = 128, VariantBoard: Type[chess.Board] = chess.Board) -> None:
         self.variant = VariantBoard
 
         self.max_fds = max_fds
-        self.lru = collections.deque()
+        self.lru: Deque[Table] = collections.deque()
         self.lru_lock = threading.Lock()
 
-        self.wdl = {}
-        self.dtz = {}
+        self.wdl: MutableMapping[str, WdlTable] = {}
+        self.dtz: MutableMapping[str, DtzTable] = {}
 
-    def _bump_lru(self, table):
+    def _bump_lru(self, table: Table) -> None:
         if self.max_fds is None:
             return
 
@@ -1490,7 +1493,7 @@ class Tablebase:
                 if len(self.lru) > self.max_fds:
                     self.lru.pop().close()
 
-    def _open_table(self, hashtable, Table, path):
+    def _open_table(self, hashtable: MutableMapping[str, Table], Table: Type[Table], path: str) -> int:
         table = Table(path, variant=self.variant)
 
         if table.key in hashtable:
@@ -1500,7 +1503,7 @@ class Tablebase:
         hashtable[table.mirrored_key] = table
         return 1
 
-    def add_directory(self, directory, *, load_wdl=True, load_dtz=True):
+    def add_directory(self, directory: str, *, load_wdl: bool = True, load_dtz: bool = True) -> int:
         """
         Adds tables from a directory.
 
@@ -1535,7 +1538,7 @@ class Tablebase:
 
         return num
 
-    def probe_wdl_table(self, board):
+    def probe_wdl_table(self, board: chess.Board) -> int:
         # Test for variant end.
         if board.is_variant_win():
             return 2
@@ -1552,13 +1555,13 @@ class Tablebase:
         try:
             table = self.wdl[key]
         except KeyError:
-            raise MissingTableError("did not find wdl table {}".format(key))
+            raise MissingTableError(f"did not find wdl table {key}")
 
         self._bump_lru(table)
 
         return table.probe_wdl_table(board)
 
-    def probe_ab(self, board, alpha, beta, threats=False):
+    def probe_ab(self, board: chess.Board, alpha: int, beta: int, threats: bool = False) -> Tuple[int, int]:
         if self.variant.captures_compulsory:
             if board.is_variant_win():
                 return 2, 2
@@ -1590,7 +1593,7 @@ class Tablebase:
         else:
             return v, 1
 
-    def sprobe_ab(self, board, alpha, beta, threats=False):
+    def sprobe_ab(self, board: chess.Board, alpha: int, beta: int, threats: bool = False) -> Tuple[int, int]:
         if chess.popcount(board.occupied_co[not board.turn]) > 1:
             v, captures_found = self.sprobe_capts(board, alpha, beta)
             if captures_found:
@@ -1622,7 +1625,7 @@ class Tablebase:
         else:
             return alpha, 3 if threats_found else 1
 
-    def sprobe_capts(self, board, alpha, beta):
+    def sprobe_capts(self, board: chess.Board, alpha: int, beta: int) -> Tuple[int, int]:
         captures_found = False
 
         for move in board.generate_legal_captures():
@@ -1642,7 +1645,7 @@ class Tablebase:
 
         return alpha, captures_found
 
-    def probe_wdl(self, board):
+    def probe_wdl(self, board: chess.Board) -> int:
         """
         Probes WDL tables for win/draw/loss-information.
 
@@ -1675,11 +1678,11 @@ class Tablebase:
         """
         # Positions with castling rights are not in the tablebase.
         if board.castling_rights:
-            raise KeyError("syzygy tables do not contain positions with castling rights: {}".format(board.fen()))
+            raise KeyError(f"syzygy tables do not contain positions with castling rights: {board.fen()}")
 
         # Validate piece count.
         if chess.popcount(board.occupied) > 7:
-            raise KeyError("syzygy tables support up to 6 (and experimentally 7) pieces, not {}: {}".format(chess.popcount(board.occupied), board.fen()))
+            raise KeyError(f"syzygy tables support up to 6 (and experimentally 7) pieces, not {chess.popcount(board.occupied)}: {board.fen()}")
 
         # Probe.
         v, _ = self.probe_ab(board, -2, 2)
@@ -1714,24 +1717,24 @@ class Tablebase:
 
         return v
 
-    def get_wdl(self, board, default=None):
+    def get_wdl(self, board: chess.Board, default: Optional[int] = None) -> Optional[int]:
         try:
             return self.probe_wdl(board)
         except KeyError:
             return default
 
-    def probe_dtz_table(self, board, wdl):
+    def probe_dtz_table(self, board: chess.Board, wdl: int) -> Tuple[int, int]:
         key = calc_key(board)
         try:
             table = self.dtz[key]
         except KeyError:
-            raise MissingTableError("did not find dtz table {}".format(key))
+            raise MissingTableError(f"did not find dtz table {key}")
 
         self._bump_lru(table)
 
         return table.probe_dtz_table(board, wdl)
 
-    def probe_dtz_no_ep(self, board):
+    def probe_dtz_no_ep(self, board: chess.Board) -> int:
         wdl, success = self.probe_ab(board, -2, 2, threats=True)
 
         if wdl == 0:
@@ -1803,7 +1806,7 @@ class Tablebase:
 
             return best
 
-    def probe_dtz(self, board):
+    def probe_dtz(self, board: chess.Board) -> int:
         """
         Probes DTZ tables for distance to zero information.
 
@@ -1911,13 +1914,13 @@ class Tablebase:
 
         return v
 
-    def get_dtz(self, board, default=None):
+    def get_dtz(self, board: chess.Board, default: Optional[int] = None) -> Optional[int]:
         try:
             return self.probe_dtz(board)
         except KeyError:
             return default
 
-    def close(self):
+    def close(self) -> None:
         """Closes all loaded tables."""
         while self.wdl:
             _, wdl = self.wdl.popitem()
@@ -1929,14 +1932,14 @@ class Tablebase:
 
         self.lru.clear()
 
-    def __enter__(self):
+    def __enter__(self) -> "Tablebase":
         return self
 
-    def __exit__(self, exc_type, exc_value, traceback):
+    def __exit__(self, exc_type: Optional[Type[BaseException]], exc_value: Optional[BaseException], traceback: Optional[TracebackType]) -> None:
         self.close()
 
 
-def open_tablebase(directory, *, load_wdl=True, load_dtz=True, max_fds=128, VariantBoard=chess.Board):
+def open_tablebase(directory: str, *, load_wdl: bool = True, load_dtz: bool = True, max_fds: Optional[int] = 128, VariantBoard: Type[chess.Board] = chess.Board) -> Tablebase:
     """
     Opens a collection of tables for probing. See
     :class:`~chess.syzygy.Tablebase`.
