@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-#
 # This file is part of the python-chess library.
 # Copyright (C) 2016-2020 Niklas Fiekas <niklas.fiekas@backscattering.de>
 #
@@ -47,7 +45,10 @@ class SuicideBoard(chess.Board):
     def _attacked_for_king(self, path: chess.Bitboard, occupied: chess.Bitboard) -> bool:
         return False
 
-    def is_check(self) -> bool:
+    def checkers_mask(self) -> chess.Bitboard:
+        return chess.BB_EMPTY
+
+    def gives_check(self, move: chess.Move) -> bool:
         return False
 
     def is_into_check(self, move: chess.Move) -> bool:
@@ -275,8 +276,8 @@ class AtomicBoard(chess.Board):
         for explosion in chess.scan_forward(explosion_radius):
             self._remove_piece_at(explosion)
 
-    def is_check(self) -> bool:
-        return not self._kings_connected() and super().is_check()
+    def checkers_mask(self) -> chess.Bitboard:
+        return chess.BB_EMPTY if self._kings_connected() else super().checkers_mask()
 
     def was_into_check(self) -> bool:
         return not self._kings_connected() and super().was_into_check()
@@ -315,6 +316,8 @@ class AtomicBoard(chess.Board):
             status &= ~chess.STATUS_NO_WHITE_KING
         else:
             status &= ~chess.STATUS_NO_BLACK_KING
+        if chess.popcount(self.checkers_mask()) <= 14:
+            status &= ~chess.STATUS_TOO_MANY_CHECKERS
         return status
 
 
@@ -360,18 +363,12 @@ class RacingKingsBoard(chess.Board):
     def reset(self) -> None:
         self.set_fen(type(self).starting_fen)
 
-    def _gives_check(self, move: chess.Move) -> bool:
-        self.push(move)
-        gives_check = self.is_check()
-        self.pop()
-        return gives_check
-
     def is_legal(self, move: chess.Move) -> bool:
-        return super().is_legal(move) and not self._gives_check(move)
+        return super().is_legal(move) and not self.gives_check(move)
 
     def generate_legal_moves(self, from_mask: chess.Bitboard = chess.BB_ALL, to_mask: chess.Bitboard = chess.BB_ALL) -> Iterator[chess.Move]:
         for move in super().generate_legal_moves(from_mask, to_mask):
-            if not self._gives_check(move):
+            if not self.gives_check(move):
                 yield move
 
     def is_variant_end(self) -> bool:
@@ -390,7 +387,7 @@ class RacingKingsBoard(chess.Board):
         # White has reached the backrank. The game is over if black can not
         # also reach the backrank on the next move. Check if there are any
         # safe squares for the king.
-        targets = chess.BB_KING_ATTACKS[black_king] & chess.BB_RANK_8
+        targets = chess.BB_KING_ATTACKS[black_king] & chess.BB_RANK_8 & ~self.occupied_co[chess.BLACK]
         return all(self.attackers_mask(chess.WHITE, target) for target in chess.scan_forward(targets))
 
     def is_variant_draw(self) -> bool:
@@ -409,7 +406,7 @@ class RacingKingsBoard(chess.Board):
     def status(self) -> chess.Status:
         status = super().status()
         if self.is_check():
-            status |= chess.STATUS_RACE_CHECK
+            status |= chess.STATUS_RACE_CHECK | chess.STATUS_TOO_MANY_CHECKERS
         if self.turn == chess.BLACK and all(self.occupied_co[co] & self.kings & chess.BB_RANK_8 for co in chess.COLORS):
             status |= chess.STATUS_RACE_OVER
         if self.pawns:
@@ -589,13 +586,7 @@ class ThreeCheckBoard(chess.Board):
         return self.remaining_checks[self.turn] <= 0 < self.remaining_checks[not self.turn]
 
     def is_irreversible(self, move: chess.Move) -> bool:
-        if super().is_irreversible(move):
-            return True
-
-        self.push(move)
-        gives_check = self.is_check()
-        self.pop()
-        return gives_check
+        return super().is_irreversible(move) or self.gives_check(move)
 
     def _transposition_key(self) -> Hashable:
         return (super()._transposition_key(),
@@ -629,22 +620,27 @@ class _CrazyhouseBoardState(Generic[CrazyhouseBoardT], chess._BoardState["Crazyh
 CrazyhousePocketT = TypeVar("CrazyhousePocketT", bound="CrazyhousePocket")
 
 class CrazyhousePocket:
+    """A Crazyhouse pocket with a counter for each piece type."""
 
     def __init__(self, symbols: Iterable[str] = "") -> None:
         self.pieces: Dict[chess.PieceType, int] = {}
         for symbol in symbols:
             self.add(chess.PIECE_SYMBOLS.index(symbol))
 
-    def add(self, pt: chess.PieceType) -> None:
-        self.pieces[pt] = self.pieces.get(pt, 0) + 1
+    def add(self, piece_type: chess.PieceType) -> None:
+        """Adds a piece of the given type to this pocket."""
+        self.pieces[piece_type] = self.pieces.get(piece_type, 0) + 1
 
-    def remove(self, pt: chess.PieceType) -> None:
-        self.pieces[pt] -= 1
+    def remove(self, piece_type: chess.PieceType) -> None:
+        """Removes a piece of the given type from this pocket."""
+        self.pieces[piece_type] -= 1
 
     def count(self, piece_type: chess.PieceType) -> int:
+        """Returns the number of pieces of the given type in the pocket."""
         return self.pieces.get(piece_type, 0)
 
     def reset(self) -> None:
+        """Clears the pocket."""
         self.pieces.clear()
 
     def __str__(self) -> str:
@@ -657,6 +653,7 @@ class CrazyhousePocket:
         return f"CrazyhousePocket('{self}')"
 
     def copy(self: CrazyhousePocketT) -> CrazyhousePocketT:
+        """Returns a copy of this pocket."""
         pocket = type(self)()
         pocket.pieces = copy.copy(self.pieces)
         return pocket
@@ -725,7 +722,7 @@ class CrazyhouseBoard(chess.Board):
         if not king_attackers:
             return ~self.occupied
         elif chess.popcount(king_attackers) == 1:
-            return chess.BB_BETWEEN[king][chess.msb(king_attackers)] & ~self.occupied
+            return chess.between(king, chess.msb(king_attackers)) & ~self.occupied
         else:
             return chess.BB_EMPTY
 
